package main;

/**
 * <h1>Evaluator</h1>
 * This class calculates the value of an expression, using the visitor pattern.
 * It visits every element of a given expression and combines them with the matching
 * arithmetical operator.
 * @version 1.0
 * @since 09.01.2023
 * @see #visit(Parser.Expression, int)
 */
public class Evaluator implements Visitor<Double> {

    private double value;

    /**
     * @see #visit(Parser.Expression, int)
     */
    @Override
    public Double visit(final Parser.BinaryOperation expression) {
        double result = expression.leftExpression.accept(this);
        double rOperand = expression.rightExpression.accept(this);
        switch (expression.operator) {
            case "+":
                result += rOperand;
                break;
            case "-":
                result -= rOperand;
                break;
            case "*":
                result *= rOperand;
                break;
            case "/":
                if (rOperand!=0) {
                    result /= rOperand;
                }
                else {
                    throw new ArithmeticException("ArithmeticError: division by zero detected");
                }
                result /= rOperand;
                break;
            case "^":
                for(double index=1.0; index<rOperand; index++) {
                    result *= result;
                }
                break;
            default:
                throw new RuntimeException("SemanticError: unknown operand found in AST: "
                + expression.operator);
        }
        return result;
    }

    /**
     * @see #visit(Parser.Expression, int)
     */
    @Override
    public Double visit(final Parser.Variable variable) {
        return value;
    }

    /**
     * @see #visit(Parser.Expression, int)
     */
    @Override
    public Double visit(final Parser.Decimal decimal) {
        return Double.parseDouble(decimal.beforeDot.digits + "." + decimal.afterDot.digits);
    }

    /**
     * @see #visit(Parser.Expression, int)
     */
    @Override
    public Double visit(final Parser.Number number) {
        return Double.parseDouble(number.digits);
    }

    /**
     * This method is used for visiting every element of an expression
     * @param ast The expression in a form of an AST(Abstract Syntax Tree)
     * @param value The int value which will be used for the variables
     * @return The final value of the expression
     */
    public double visit(final Parser.Expression ast, int value) {
        this.value = value;
        return ast.accept(this);
    }
}
