package main;

import Exceptions.LexerException;

import java.util.LinkedList;
import java.util.List;

/**
 * <h1>Lexer</h1>
 * The Lexer class crates a List of Token objects out of a String
 *
 * @version 1.0
 * @since 09.01.2023
 * @see TokenType
 * @see Token
 * @see #lex(String)
 */
public class Lexer {

    /**
     * enum TokenType that can either be a number, variable or a special character (like * or / etc.)
     * @see Lexer
     */
    public static enum TokenType {
        NUMBER, VARIABLE, SPECIAL
    }

    /**
     * tokens with a TokenType that represents their type and a String with the actual data
     * has for both values a getter method and overrides toString with its fitting values
     * @see Lexer
     */
    public static class Token {
        protected TokenType type;
        protected String data;
        public Token(TokenType type, String data) {
            this.type = type;
            this.data = data;
        }

        public TokenType getType() {
            return type;
        }

        public String getData() {
            return data;
        }

        @Override
        public String toString() {
            return String.format("type: %s, data: %s", type, data);
        }
    }

    /**
     * creates a list of tokens from a given String
     * @param input The String with the input
     * @return The List of Token objects
     * @throws LexerException If the String was empty, or it couldn't create any result out of it
     * @see Lexer
     */
    public static List<Token> lex(String input) throws Exception{
        if(input.isEmpty()) {
            throw new LexerException("SyntaxError: lexer received empty String");
        }

        //the list that will be returned at the end
        List<Token> result = new LinkedList<Token>();
        //for-loop that iterates through each character of the String
        for(int index=0; index < input.length();) {

            char current = input.charAt(index);
            //checks if a character is a digit, then creates a token with one or multiple
            //numbers with TokenType=NUMBER
            if(Character.isDigit(current)) {
                int endIndex = index;
                if(endIndex<input.length()) {
                    while(Character.isDigit(input.charAt(endIndex)) && endIndex<input.length()) {
                        endIndex += 1;
                        if(endIndex==input.length()) {
                            break;
                        }
                        if(input.charAt(endIndex) == '.') {
                            endIndex+=1;
                        }
                    }
                }
                Token token = new Token(TokenType.NUMBER, input.substring(index,endIndex));
                result.add(token);
                index += endIndex - index;
            }
            //checks if a character is a letter, then creates for each following character a separate token
            //with TokenType=VARIABLE
            else if(Character.isLetter(current)){
                int iterator = index;
                while(Character.isLetter(input.charAt(iterator)) && iterator<input.length()) {
                    Token token = new Token(TokenType.VARIABLE, String.valueOf(input.charAt(iterator)));
                    result.add(token);
                    iterator += 1;
                    if(!(iterator<input.length())) {
                        break;
                    }
                }
                index += iterator - index;
            }
            //whitespaces are ignored
            else if(Character.isWhitespace(current)) {
                index++;
            }
            //every char of the string that isn't a digit, letter or whitespace will be
            //used for creating a token with TokenType=SPECIAL
            else {
                Token token = new Token(TokenType.SPECIAL, String.valueOf(current));
                result.add(token);
                index++;
            }
        }
        if(result.isEmpty()) {
            throw new LexerException("SyntaxError: lexer couldn't create any tokens out of the input");
        }
        return result;
    }
}
