package main;

import Exceptions.ParserException;

import java.util.LinkedList;
import java.util.List;

/**
 * <h1>Parser</h1>
 * The Parser class parses a list of Tokens into an Expression that represents an AST
 *
 * @version 1.0
 * @since 09.01.2023
 * @see #parse(List)
 * @see Expression
 */
public class Parser {

    /**
     * This abstract class is only used for heredity
     * It represents an expression in the form of an Abstract Syntax Tree(AST)
     * A method 'accept' is implemented, it is used for allowing other classes to interact with it
     * through the visiting pattern
     * @see BinaryOperation
     * @see Variable
     * @see Value
     */
    public abstract class Expression {
        public abstract <T> T accept(Visitor<T> visitor);
    }

    /**
     * This class represents a binary operation
     * it contains a left expression, an operator, and a right expression
     * if this operation is capsuled by brackets, the value of 'capsuled' will be true, else it will be false
     * @see Variable
     * @see Value
     */
    public class BinaryOperation extends Expression {
        Expression leftExpression;
        String operator;
        Expression rightExpression;
        Boolean capsuled;
        public BinaryOperation(Expression leftExpression, String operator, Expression rightExpression, Boolean capsuled) {
            this.leftExpression = leftExpression;
            this.operator = operator;
            this.rightExpression = rightExpression;
            this.capsuled = capsuled;
        }

        @Override
        public <T> T accept(Visitor<T> visitor) {
            return visitor.visit(this);
        }
    }

    /**
     * This class represents a variable, with it holds a String 'variableName' that contains a character
     * of the alphabet
     */
    public class Variable extends Expression {
        String variableName;
        public Variable(String i) {
            variableName = i;
        }
        public <T> T accept(Visitor<T> visitor) {
            return visitor.visit(this);
        }
    }

    /**
     * This abstract class is only used for heredity
     * A value can either be a number or a decimal
     * @see Decimal
     * @see Number
     */
    private abstract class Value extends Expression {

    }

    /**
     * This class represents a number, it contains a String 'digits' that holds it value
     */
    public class Number extends Value {
        String digits;
        public Number(String i) {
            this.digits = i;
        }
        public <T> T accept(Visitor<T> visitor) {
            return visitor.visit(this);
        }
    }

    /**
     * This class represents a decimal, it contains two numbers: the one before- and the one after the dot
     * @see Number
     */
    public class Decimal extends Value {
        public Number beforeDot;
        public Number afterDot;

        public Decimal(Number i1, Number i2) {
            this.beforeDot = i1;
            this.afterDot = i2;
        }
        public <T> T accept(Visitor<T> visitor) {
            return visitor.visit(this);
        }
    }

    /**
     * starting method of the parser that parses a list of tokens into an AST (Abstract Syntax Tree)
     * @param list The list of tokens
     * @return Expression Returns the Expression object that represents the AST
     * @throws ParserException If the list of tokens was empty or not all tokens could be used in
     * the process due to wrong syntax/semantic
     */
    public Expression parse(List<Lexer.Token> list) throws ParserException {
        if(list.isEmpty()) {
            throw new ParserException("empty token list");
        }

        List<Lexer.Token> ts = new LinkedList<>(list);
        Expression ast = parseExpression(ts);

        if(!ts.isEmpty()) {
            throw new ParserException("SyntaxError: " + ts.size() + " token(s) left");
        }
        return ast;
    }

    /**
     * parses a list of tokens into an expression (maybe with recursive calls through parseBinaryOperation(...)
     * @param ts The list of tokens
     * @return The expression object that got parsed
     * @throws ParserException If the list of tokens was empty
     * @see #parseBinaryOperation(List) 
     * @see #parseVariable(List)
     * @see #parseDecimal(String) 
     * @see #parseVariable(List) 
     */
    private Expression parseExpression(List<Lexer.Token> ts) throws ParserException {
        if(ts.isEmpty()) {
            throw new ParserException("SyntaxError: empty token list");
        }
        Expression ast = parseBinaryOperation(ts);
            if(ast==null) {
                ast = parseVariable(ts);
                if(ast==null) {
                    ast = parseValue(ts);
                }
                if(ast==null) {
                    throw new ParserException("SyntaxError: invalid character");
                }
            }
        return ast;
    }

    /**
     * checks if there are multiple tokens left, which indicates that they build up a BinaryOperation
     * It splits up the tokens into a left and right expression and with a recursive call of parseExpression
     * each part will be determined, until all BinaryOperations consist of finite left and right expressions
     * @param ts The list of tokens
     * @return null If the list does not contain a BinaryOperation <br> or: BinaryOperation The crated BinaryOP
     * @throws ParserException If the list of tokens was empty
     * @see #parseBrackets(List)
     * */
    private BinaryOperation parseBinaryOperation(List<Lexer.Token> ts) throws ParserException {
        if(ts.isEmpty()) {
            throw new ParserException("SyntaxError: empty token list");
        }
        if(ts.size()==1) {
            return null;
        }
        //Index will be determined in the next if/else statement
        //It will indicate the position of the operator that separates the left- and right expression
        int index;
        int tokensInBracket = parseBrackets(ts);
        boolean capsuled;
        //If the result of parseBrackets(ts)+1 is the same as ts.size the current BinaryOperation is capsuled
        if(tokensInBracket+1==ts.size()) {
            //removes the tokens with the capsuling brackets
            ts.remove(tokensInBracket);
            ts.remove(0);
            capsuled = true;
            //Checks again for brackets
            tokensInBracket = parseBrackets(ts);
            //If no bracket is following, the operator needs to be at position 1 of ts
            if(tokensInBracket==-1) {
                index = 1;
            }
            else {
                //If brackets are following, the operator is at position tokensInBracket+1
                index = tokensInBracket+1;
            }
        }
        //If tokensInBracket is -1 this expression is not capsuled and the operator needs to be at position 1 of ts
        else if (tokensInBracket==-1) {
            capsuled = false;
            index = 1;
        }
        //If both other cases were not fulfilled an inner expression is capsuled
        //Then this BinaryOperation is not capsuled, no brackets need to be removed and the operator is at
        //position tokensInBracket+1
        else {
            capsuled = false;
            index = tokensInBracket+1;
        }

        //Checks if the operator of this BinaryOperation is permitted
        parseOperator(ts.get(index).getData());
        List<Lexer.Token> leftList = new LinkedList<>();
        //Splits the token list ts into a left and right expression
        for(int iterator = 0; iterator<index; iterator++) {
            leftList.add(ts.remove(0));
        }
        Expression leftExpression = parseExpression(leftList);
        //Operator is now at position 0
        String operator = ts.remove(0).getData();
        Expression rightExpression = parseExpression(ts);
        return new BinaryOperation(leftExpression, operator, rightExpression, capsuled);
    }


    /**
     * checks if the first token is a bracket
     * @param ts The List of tokens
     * @return int The position of the token with the closing bracket or -1 if the first token wasn't a bracket
     * @throws ParserException If the list of tokens was empty, a bracket never gets closed or there isn't any
     * value between the brackets
     * @see #parseCharacter(String, String)
     */
    private int parseBrackets(List<Lexer.Token> ts) throws ParserException{
        if(ts.isEmpty()) {
            throw new ParserException("SyntaxError: empty token list");
        }
        if(parseCharacter(ts.get(0).getData(), "(")) {
            int index;
            int lBrackets = 0;
            int rBrackets = 0;
            boolean found = false;
            //Iterates through every token, until the closing bracket got found
            for(index=1; index<ts.size(); index++) {
                if(parseCharacter(ts.get(index).getData(), "(")) {
                    lBrackets += 1;
                }
                else if(parseCharacter(ts.get(index).getData(), ")")) {
                    rBrackets += 1;
                    //If more ')' got found then '(' the bracket got closed correctly and the loop can be interrupted
                    if(rBrackets>lBrackets) {
                        found = true;
                        break;
                    }
                }
            }
            //If no matching closing bracket was found a ParserException will be thrown
            if(!found) {
                throw new ParserException("SyntaxError: bracket never got closed");
            }
            //If the closing bracket was placed immediately after the opening bracket a ParserException will be thrown
            if(index==1) {
                throw new ParserException("SyntaxError: no value inside brackets");
            }
            //Returns the index of the closing bracket
            return index;
        }
        return -1;
    }


    /**
     * Checks if a String only contains an allowed operator
     * @param operator The string that represents the operator
     * @throws ParserException If the operator token was too big, small or consisted of invalid characters
     * @see #parseCharacter(String, String) 
     */
    private void parseOperator(String operator) throws ParserException {
        if(operator.isEmpty()) {
            throw new ParserException("RuntimeError: empty String instead of an operator (+, -, *, /, ^");
        }
        if(operator.length()>1) {
            throw new ParserException("SyntaxError: invalid length for an operator: " + operator);
        }
        if(!parseCharacter(operator, "+-*/^")) {
            throw new ParserException("SyntaxError: invalid operator character: " + operator);
        }
    }

    /**
     * Checks if a token is a variable:
     * @param ts List of tokens
     * @return null It the token isn't a variable <br>or: Variable A new Variable that gets created
     * with the received token
     * @throws ParserException If the list of tokens was empty or the size of the token was too big
     * or too small for a variable
     */
    private Variable parseVariable(List<Lexer.Token> ts) throws ParserException {
        if (ts.isEmpty()) {
            throw new ParserException("SyntaxError: empty token list");
        }
        //If the TokenType is != VARIABLE the token is not a variable and null gets returned
        if (ts.get(0).getType() != Lexer.TokenType.VARIABLE) {
            return null;
        }
        if (ts.get(0).getData().length()>1) {
            throw new ParserException("Invalid size for a variable!");
        }

        String data = ts.remove(0).getData();
        //If the token is empty, throw a ParserException
        if (data.isEmpty()) {
            throw new ParserException("SyntaxError: empty token");
        }
        //Checks if the variable is a character of the alphabet, if not the token is not valid
        if(!parseCharacter(data,"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz")) {
            throw new ParserException("Invalid variable character: " + data);
        }
        return new Variable(data);
    }


    /**
     * Checks if a token is a value and if so creates and returns a new Number or Decimal object
     * @param ts The List of the tokens
     * @return null If the token isn't a value/number <br>or: Value The created Number or Decimal Object
     * @throws ParserException If the list of tokens is empty or
     * @see #parseDecimal(String)
     * @see #parseNumber(String, boolean)
     */
    private Value parseValue(List<Lexer.Token> ts) throws ParserException {
        if (ts.isEmpty()) {
            throw new ParserException("SyntaxError: empty token list");
        }
        if (ts.get(0).getType() != Lexer.TokenType.NUMBER) {
            return null;
        }
        String data = ts.remove(0).getData();
        if (data.isEmpty()) {
            throw new ParserException("SyntaxError: empty token");
        }

        if(data.contains(".")) {
            return parseDecimal(data);
        }
        return parseNumber(data, false);
    }


    /**
     * parses a String into a Decimal object that consists of two numbers
     * @param data The String that contains the value
     * @return Decimal Returns the new Decimal object
     * @see #parseVariable(List)
     */
    private Decimal parseDecimal(String data) throws ParserException {
        if(data.isEmpty()) {
            throw new ParserException("SyntaxError: empty data");
        }
        //Separates the String into the number before- and after the dot
        int dot = data.indexOf('.');
        String number1 = data.substring(0,dot);
        String number2 = data.substring(dot+1);
        Number beforeDot = (Number) parseNumber(number1, true);
        Number afterDot = (Number) parseNumber(number2, false);
        return new Decimal(beforeDot, afterDot);
    }

    /**
     * parses a String into a Number object
     * @param data The String that contains the value
     * @param afterDot If this Number will be used after a dot in a Decimal object
     * @return Number Returns the new Number object
     * @see #parseDigit(String) 
     * @see #parseDigitWithoutZero(String) 
     */
    private Number parseNumber(String data, boolean afterDot) throws ParserException{
        if (data.isEmpty()) {
            throw new ParserException("RuntimeException: empty token");
        }
        //if the number only contains the digit 0, it can be returned already
        if (data.startsWith("0") && data.length() == 1) {
            return new Number(data);
        }
        //Checks if all digits of the number are permitted
        //If afterDot is false, the first digit is not allowed to be 0
        if(!afterDot) {
            parseDigitWithoutZero(data.substring(0,1));

            if (data.length()>1) {
                parseDigit(data.substring(1));
            }
        }
        //Checks if all digits of the number are permitted
        else {
            parseDigit(data);
        }
        return new Number(data);
    }

    /**
     * checks if a String only contains numbers(including zero)
     * @param data The String that gets checked
     * @exception ParserException If the data is empty or does contain invalid characters
     * @see #parseCharacter(String, String) 
     */
    private void parseDigit(String data) throws ParserException {
        for(int index=0; index<data.length(); index++) {
            if(data.isEmpty()) {
                throw new ParserException("RuntimeError: empty String instead of a digit(0-9)");
            }
            String character = Character.toString(data.charAt(index));
            if(!parseCharacter(character, "0123456789")) {
                throw new ParserException("SyntaxError: unexpected character: " + character);
            }
        }
    }

    /**
     * checks if a String only contains numbers(excluding zero)
     * @param data The String that gets checked
     * @exception ParserException If the data is empty or does contain invalid characters
     * @see #parseCharacter(String, String) 
     */
    private void parseDigitWithoutZero(String data) throws ParserException {
        if(data.isEmpty()) {
            throw new ParserException("RuntimeError: empty String instead of a digit(1-9)");
        }
        for(int index=0; index<data.length(); index++) {
            String character = Character.toString(data.charAt(index));
            if(!parseCharacter(character, "123456789")) {
                throw new ParserException("SyntaxError: unexpected character: " + character);
            }
        }
    }

    /**
     * checks if a certain string can be found in a String of allowed characters
     * @param data The String
     * @param allowedCharacters The String of allowed characters
     * @return boolean Returns if data was found in allowedCharacters or not
     * @exception ParserException If the data is empty
     */
    private boolean parseCharacter(String data, String allowedCharacters) throws ParserException {
        if(data.isEmpty()) {
            throw new ParserException("RuntimeError: empty String");
        }
        return (allowedCharacters.contains(data));
    }
}
